@@ -1,29 +1,18 @@
-#include <sys/types.h>
-#include <sys/socket.h>
-#include <sys/un.h>
-#include <arpa/inet.h>
-#include <sys/stat.h>
-
-#include <inttypes.h>
-#include <limits.h>
-#include <stdarg.h>
-#include <stdio.h>
-#include <string.h>
-#include <unistd.h>
-
 #include "btpd.h"
-#include "active.h"
-#include "tracker_req.h"
+
+#include <sys/un.h>
+#include <iobuf.h>
 
 struct cli {
     int sd;
-    struct event read;
+    struct fdev read;
 };
 
-static struct event m_cli_incoming;
+static int m_listen_sd;
+static struct fdev m_cli_incoming;
 
 static int
-write_buffer(struct cli *cli, struct io_buffer *iob)
+write_buffer(struct cli *cli, struct iobuf *iob)
 {
     int err = 0;
     if (!iob->error) {
@@ -32,106 +21,107 @@ write_buffer(struct cli *cli, struct io_buffer *iob)
         err = write_fully(cli->sd, iob->buf, iob->off);
     } else
         btpd_err("Out of memory.\n");
-    buf_free(iob);
+    iobuf_free(iob);
     return err;
 }
 
 static int
 write_code_buffer(struct cli *cli, enum ipc_err code)
 {
-    struct io_buffer iob = buf_init(16);
-    buf_print(&iob, "d4:codei%uee", code);
+    struct iobuf iob = iobuf_init(16);
+    iobuf_print(&iob, "d4:codei%uee", code);
     return write_buffer(cli, &iob);
 }
 
 static int
 write_add_buffer(struct cli *cli, unsigned num)
 {
-    struct io_buffer iob = buf_init(32);
-    buf_print(&iob, "d4:codei%ue3:numi%uee", IPC_OK, num);
+    struct iobuf iob = iobuf_init(32);
+    iobuf_print(&iob, "d4:codei%ue3:numi%uee", IPC_OK, num);
     return write_buffer(cli, &iob);
 }
 
 static void
-write_ans(struct io_buffer *iob, struct tlib *tl, enum ipc_tval val)
+write_ans(struct iobuf *iob, struct tlib *tl, enum ipc_tval val)
 {
     enum ipc_tstate ts = IPC_TSTATE_INACTIVE;
     switch (val) {
     case IPC_TVAL_CGOT:
-        buf_print(iob, "i%dei%llde", IPC_TYPE_NUM,
+        iobuf_print(iob, "i%dei%llde", IPC_TYPE_NUM,
             tl->tp == NULL ? tl->content_have : (long long)cm_content(tl->tp));
         return;
     case IPC_TVAL_CSIZE:
-        buf_print(iob, "i%dei%llde", IPC_TYPE_NUM, tl->content_size);
+        iobuf_print(iob, "i%dei%llde", IPC_TYPE_NUM,
+            (long long)tl->content_size);
         return;
     case IPC_TVAL_PCCOUNT:
         if (tl->tp == NULL)
-            buf_print(iob, "i%dei%de", IPC_TYPE_ERR, IPC_ETINACTIVE);
+            iobuf_print(iob, "i%dei%de", IPC_TYPE_ERR, IPC_ETINACTIVE);
         else
-            buf_print(iob, "i%dei%lue", IPC_TYPE_NUM,
+            iobuf_print(iob, "i%dei%lue", IPC_TYPE_NUM,
                 (unsigned long)tl->tp->npieces);
         return;
     case IPC_TVAL_PCGOT:
         if (tl->tp == NULL)
-            buf_print(iob, "i%dei%de", IPC_TYPE_ERR, IPC_ETINACTIVE);
+            iobuf_print(iob, "i%dei%de", IPC_TYPE_ERR, IPC_ETINACTIVE);
         else
-            buf_print(iob, "i%dei%lue", IPC_TYPE_NUM,
+            iobuf_print(iob, "i%dei%lue", IPC_TYPE_NUM,
                 (unsigned long)cm_pieces(tl->tp));
         return;
     case IPC_TVAL_PCSEEN:
         if (tl->tp == NULL)
-            buf_print(iob, "i%dei%de", IPC_TYPE_NUM, 0);
+            iobuf_print(iob, "i%dei%de", IPC_TYPE_NUM, 0);
         else {
             unsigned long pcseen = 0;
             for (unsigned long i = 0; i < tl->tp->npieces; i++)
                 if (tl->tp->net->piece_count[i] > 0)
                     pcseen++;
-            buf_print(iob, "i%dei%lue", IPC_TYPE_NUM, pcseen);
+            iobuf_print(iob, "i%dei%lue", IPC_TYPE_NUM, pcseen);
         }
         return;
     case IPC_TVAL_RATEDWN:
-        buf_print(iob, "i%dei%lue", IPC_TYPE_NUM,
+        iobuf_print(iob, "i%dei%lue", IPC_TYPE_NUM,
             tl->tp == NULL ? 0UL : tl->tp->net->rate_dwn / RATEHISTORY);
         return;
     case IPC_TVAL_RATEUP:
-        buf_print(iob, "i%dei%lue", IPC_TYPE_NUM,
+        iobuf_print(iob, "i%dei%lue", IPC_TYPE_NUM,
             tl->tp == NULL ? 0UL : tl->tp->net->rate_up / RATEHISTORY);
         return;
     case IPC_TVAL_SESSDWN:
-        buf_print(iob, "i%dei%llde", IPC_TYPE_NUM,
+        iobuf_print(iob, "i%dei%llde", IPC_TYPE_NUM,
             tl->tp == NULL ? 0LL : tl->tp->net->downloaded);
         return;
     case IPC_TVAL_SESSUP:
-        buf_print(iob, "i%dei%llde", IPC_TYPE_NUM,
+        iobuf_print(iob, "i%dei%llde", IPC_TYPE_NUM,
             tl->tp == NULL ? 0LL : tl->tp->net->uploaded);
         return;
     case IPC_TVAL_DIR:
         if (tl->dir != NULL)
-            buf_print(iob, "i%de%d:%s", IPC_TYPE_STR, (int)strlen(tl->dir),
+            iobuf_print(iob, "i%de%d:%s", IPC_TYPE_STR, (int)strlen(tl->dir),
                 tl->dir);
         else
-            buf_print(iob, "i%dei%de", IPC_TYPE_ERR, IPC_EBADTENT);
+            iobuf_print(iob, "i%dei%de", IPC_TYPE_ERR, IPC_EBADTENT);
         return;
     case IPC_TVAL_NAME:
         if (tl->name != NULL)
-            buf_print(iob, "i%de%d:%s", IPC_TYPE_STR, (int)strlen(tl->name),
+            iobuf_print(iob, "i%de%d:%s", IPC_TYPE_STR, (int)strlen(tl->name),
                 tl->name);
         else
-            buf_print(iob, "i%dei%de", IPC_TYPE_ERR, IPC_EBADTENT);
+            iobuf_print(iob, "i%dei%de", IPC_TYPE_ERR, IPC_EBADTENT);
         return;
     case IPC_TVAL_IHASH:
-        buf_print(iob, "i%de20:", IPC_TYPE_BIN);
-        buf_write(iob, tl->hash, 20);
+        iobuf_print(iob, "i%de20:", IPC_TYPE_BIN);
+        iobuf_write(iob, tl->hash, 20);
         return;
     case IPC_TVAL_NUM:
-        buf_print(iob, "i%dei%ue", IPC_TYPE_NUM, tl->num);
+        iobuf_print(iob, "i%dei%ue", IPC_TYPE_NUM, tl->num);
         return;
     case IPC_TVAL_PCOUNT:
-        buf_print(iob, "i%dei%ue", IPC_TYPE_NUM,
+        iobuf_print(iob, "i%dei%ue", IPC_TYPE_NUM,
             tl->tp == NULL ? 0 : tl->tp->net->npeers);
         return;
     case IPC_TVAL_STATE:
-        buf_print(iob, "i%de", IPC_TYPE_NUM);
+        iobuf_print(iob, "i%de", IPC_TYPE_NUM);
         if (tl->tp != NULL) {
             switch (tl->tp->state) {
             case T_STARTING:
@@ -146,26 +136,31 @@ write_ans(struct io_buffer *iob, struct tlib *tl, enum ipc_tval val)
             case T_LEECH:
                 ts = IPC_TSTATE_LEECH;
                 break;
+            case T_GHOST:
+                break;
             }
         }
-        buf_print(iob, "i%de", ts);
+        iobuf_print(iob, "i%de", ts);
         return;
     case IPC_TVAL_TOTDWN:
-        buf_print(iob, "i%dei%llde", IPC_TYPE_NUM, tl->tot_down +
+        iobuf_print(iob, "i%dei%llde", IPC_TYPE_NUM, tl->tot_down +
             (tl->tp == NULL ? 0 : tl->tp->net->downloaded));
         return;
     case IPC_TVAL_TOTUP:
-        buf_print(iob, "i%dei%llde", IPC_TYPE_NUM, tl->tot_up +
+        iobuf_print(iob, "i%dei%llde", IPC_TYPE_NUM, tl->tot_up +
             (tl->tp == NULL ? 0 : tl->tp->net->uploaded));
         return;
     case IPC_TVAL_TRERR:
-        buf_print(iob, "i%dei%ue", IPC_TYPE_NUM,
-            tl->tp == NULL ? 0 : tr_errors(tl->tp));
+        iobuf_print(iob, "i%dei%ue", IPC_TYPE_NUM, 0);
+        return;
+    case IPC_TVAL_TRGOOD:
+        iobuf_print(iob, "i%dei%de", IPC_TYPE_NUM,
+            tl->tp == NULL ? 0 : tr_good_count(tl->tp));
         return;
     case IPC_TVALCOUNT:
         break;
     }
-    buf_print(iob, "i%dei%de", IPC_TYPE_ERR, IPC_ENOKEY);
+    iobuf_print(iob, "i%dei%de", IPC_TYPE_ERR, IPC_ENOKEY);
 }
 
 static int
@@ -177,7 +172,7 @@ cmd_tget(struct cli *cli, int argc, const char *args)
     size_t nkeys;
     const char *keys, *p;
     enum ipc_tval *opts;
-    struct io_buffer iob;
+    struct iobuf iob;
 
     if ((keys = benc_dget_lst(args, "keys")) == NULL)
         return IPC_COMMERR;
@@ -189,21 +184,22 @@ cmd_tget(struct cli *cli, int argc, const char *args)
     for (int i = 0; i < nkeys; i++)
         opts[i] = benc_int(p, &p);
 
-    iob = buf_init(1 << 15);
-    buf_swrite(&iob, "d4:codei0e6:resultl");
+    iob = iobuf_init(1 << 15);
+    iobuf_swrite(&iob, "d4:codei0e6:resultl");
     p = benc_dget_any(args, "from");
     if (benc_isint(p)) {
         enum ipc_twc from = benc_int(p, NULL);
-        struct tlib *tlv[tlib_count()];
-        tlib_put_all(tlv);
-        for (int i = 0; i < sizeof(tlv) / sizeof(tlv[0]); i++) {
-            if ((from == IPC_TWC_ALL ||
-                    (tlv[i]->tp == NULL && from == IPC_TWC_INACTIVE) ||
-                    (tlv[i]->tp != NULL && from == IPC_TWC_ACTIVE))) {
-                buf_swrite(&iob, "l");
+        struct htbl_iter it;
+        struct tlib *tl;
+        for (tl = tlib_iter_first(&it); tl != NULL; tl = tlib_iter_next(&it)) {
+            if (!torrent_haunting(tl) && (
+                    from == IPC_TWC_ALL ||
+                    (!torrent_active(tl) && from == IPC_TWC_INACTIVE) ||
+                    (torrent_active(tl) && from == IPC_TWC_ACTIVE))) {
+                iobuf_swrite(&iob, "l");
                 for (int k = 0; k < nkeys; k++)
-                    write_ans(&iob, tlv[i], opts[k]);
-                buf_swrite(&iob, "e");
+                    write_ans(&iob, tl, opts[k]);
+                iobuf_swrite(&iob, "e");
             }
         }
     } else if (benc_islst(p)) {
@@ -214,20 +210,20 @@ cmd_tget(struct cli *cli, int argc, const char *args)
             else if (benc_isstr(p) && benc_strlen(p) == 20)
                 tl = tlib_by_hash(benc_mem(p, NULL, NULL));
             else {
-                buf_free(&iob);
+                iobuf_free(&iob);
                 free(opts);
                 return IPC_COMMERR;
             }
-            if (tl != NULL) {
-                buf_swrite(&iob, "l");
+            if (tl != NULL && !torrent_haunting(tl)) {
+                iobuf_swrite(&iob, "l");
                 for (int i = 0; i < nkeys; i++)
                     write_ans(&iob, tl, opts[i]);
-                buf_swrite(&iob, "e");
+                iobuf_swrite(&iob, "e");
             } else
-                buf_print(&iob, "i%de", IPC_ENOTENT);
+                iobuf_print(&iob, "i%de", IPC_ENOTENT);
         }
     }
-    buf_swrite(&iob, "ee");
+    iobuf_swrite(&iob, "ee");
     free(opts);
     return write_buffer(cli, &iob);
 }
@@ -260,10 +256,15 @@ cmd_add(struct cli *cli, int argc, const char *args)
     content[csize] = '\0';
 
     tl = tlib_by_hash(mi_info_hash(mi, hash));
-    if (tl != NULL)
+    if (tl != NULL && !torrent_haunting(tl))
         return write_code_buffer(cli, IPC_ETENTEXIST);
-    tl = tlib_add(hash, mi, mi_size, content,
-        benc_dget_str(args, "name", NULL));
+    if (tl != NULL) {
+        tl = tlib_readd(tl, hash, mi, mi_size, content,
+            benc_dget_str(args, "name", NULL));
+    } else {
+        tl = tlib_add(hash, mi, mi_size, content,
+            benc_dget_str(args, "name", NULL));
+    }
     return write_add_buffer(cli, tl->num);
 }
 
@@ -282,7 +283,7 @@ cmd_del(struct cli *cli, int argc, const char *args)
     else
         return IPC_COMMERR;
 
-    if (tl == NULL)
+    if (tl == NULL || torrent_haunting(tl))
         ret = write_code_buffer(cli, IPC_ENOTENT);
     else {
         ret = write_code_buffer(cli, IPC_OK);
@@ -312,9 +313,9 @@ cmd_start(struct cli *cli, int argc, const char *args)
     else
         return IPC_COMMERR;
 
-    if (tl == NULL)
+    if (tl == NULL || torrent_haunting(tl))
         code = IPC_ENOTENT;
-    else if (tl->tp != NULL)
+    else if (!torrent_startable(tl))
         code = IPC_ETACTIVE;
     else
         if ((code = torrent_start(tl)) == IPC_OK)
@@ -336,9 +337,9 @@ cmd_stop(struct cli *cli, int argc, const char *args)
     else
         return IPC_COMMERR;
 
-    if (tl == NULL)
+    if (tl == NULL || torrent_haunting(tl))
         return write_code_buffer(cli, IPC_ENOTENT);
-    else if (tl->tp == NULL)
+    else if (!torrent_active(tl))
         return write_code_buffer(cli, IPC_ETINACTIVE);
     else  {
         // Stopping a torrent may trigger exit so we need to reply before.
@@ -352,12 +353,11 @@ cmd_stop(struct cli *cli, int argc, const char *args)
 static int
 cmd_stop_all(struct cli *cli, int argc, const char *args)
 {
-    struct torrent *tp;
+    struct torrent *tp, *next;
     int ret = write_code_buffer(cli, IPC_OK);
     active_clear();
-    BTPDQ_FOREACH(tp, torrent_get_all(), entry)
-        if (tp->state != T_STOPPING)
-            torrent_stop(tp, 0);
+    BTPDQ_FOREACH_MUTABLE(tp, torrent_get_all(), entry, next)
+        torrent_stop(tp, 0);
     return ret;
 }
 
@@ -366,11 +366,8 @@ cmd_die(struct cli *cli, int argc, const char *args)
 {
     int err = write_code_buffer(cli, IPC_OK);
     if (!btpd_is_stopping()) {
-        int grace_seconds = -1;
-        if (argc == 1 && benc_isint(args))
-            grace_seconds = benc_int(args, NULL);
         btpd_log(BTPD_L_BTPD, "Someone wants me dead.\n");
-        btpd_shutdown(grace_seconds);
+        btpd_shutdown();
     }
     return err;
 }
@@ -431,10 +428,10 @@ cli_read_cb(int sd, short type, void *arg)
         goto error;
 
     free(msg);
-    btpd_ev_add(&cli->read, NULL);
     return;
 
 error:
+    btpd_ev_del(&cli->read);
     close(cli->sd);
     free(cli);
     if (msg != NULL)
@@ -458,8 +455,14 @@ client_connection_cb(int sd, short type, void *arg)
 
     struct cli *cli = btpd_calloc(1, sizeof(*cli));
     cli->sd = nsd;
-    event_set(&cli->read, cli->sd, EV_READ, cli_read_cb, cli);
-    btpd_ev_add(&cli->read, NULL);
+    btpd_ev_new(&cli->read, cli->sd, EV_READ, cli_read_cb, cli);
+}
+
+void
+ipc_shutdown(void)
+{
+    btpd_ev_del(&m_cli_incoming);
+    close(m_listen_sd);
 }
 
 void
@@ -489,7 +492,6 @@ ipc_init(void)
     listen(sd, 4);
     set_nonblocking(sd);
 
-    event_set(&m_cli_incoming, sd, EV_READ | EV_PERSIST,
-        client_connection_cb, NULL);
-    btpd_ev_add(&m_cli_incoming, NULL);
+    btpd_ev_new(&m_cli_incoming, sd, EV_READ, client_connection_cb, NULL);
+    m_listen_sd = sd;
 }
